﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using OpenNETCF.AppSettings;
using System.Globalization;

namespace MSIS
{

    public sealed class AppSettings
    {
        static readonly string AppPath = System.IO.Path.GetDirectoryName(
              System.Reflection.Assembly.GetExecutingAssembly().GetName().CodeBase);
        static readonly SettingsFile settings = new SettingsFile(AppPath + @"\AppSettings.xml", true);
        static SettingGroup group;

        static AppSettings()
        {
            if (settings.Groups.Count > 0)
            {
                group = settings.Groups["Configuration"];
            }
        }

        AppSettings() { }

        #region Language helping methods
        public static List<CultureInfo> AvailableLanguages()
        {
            List<CultureInfo> langs = new List<CultureInfo>();
            langs.Add(CultureInfo.GetCultureInfo("en"));
            langs.Add(CultureInfo.GetCultureInfo("nb-NO"));
            return langs;
        }
        public static CultureInfo CultureFromName(string langname)
        {
            if (langname.ToLower().Contains("bokmål")) return CultureInfo.GetCultureInfo("nb-NO");
            if (langname.ToLower().Contains("english")) return CultureInfo.GetCultureInfo("en");
            else return null;
        }
        private static CultureInfo GetDefaultLanguage()
        {
            if (CultureInfo.CurrentCulture.ToString().Equals(CultureInfo.GetCultureInfo("nb-NO").ToString())) return CultureInfo.GetCultureInfo("nb-NO");
            return CultureInfo.GetCultureInfo("en");
        }
        #endregion



        public static void Create()
        {
            if (!ConfigFileExists)
            {
                settings.Groups.Add("Configuration");
                settings.Groups["Configuration"].Settings.Add("WebServiceUrl", "http://msis.idi.ntnu.no/krypton/");
                settings.Groups["Configuration"].Settings.Add("MapUrl", "http://msis.idi.ntnu.no/maps/");
                settings.Groups["Configuration"].Settings.Add("ShowHints", true);
                settings.Groups["Configuration"].Settings.Add("RememberUsernameAndPassword", false);
                settings.Groups["Configuration"].Settings.Add("Language", GetDefaultLanguage().ToString());
                settings.Groups["Configuration"].Settings.Add("Courses", "all");
                settings.Groups["Configuration"].Settings.Add("AnnouncementsDateFilter", 0);
                settings.Save();
            }
            group = settings.Groups["Configuration"];
        }

        public static bool ConfigFileExists
        {
            get
            {
                if (group != null)
                {
                    return true;
                }
                return false;
            }
        }

        public static string WebServiceUrl
        {
            get { return group.Settings["WebServiceUrl"].Value.ToString(); }
            set
            {
                group.Settings["WebServiceUrl"].Value = value;
                settings.Save();
            }
        }

        public static string MapUrl
        {
            get { return group.Settings["MapUrl"].Value.ToString(); }
            set
            {
                group.Settings["MapUrl"].Value = value;
                settings.Save();
            }
        }

        public static string[] Courses
        {
            get 
            {
                if (group.Settings["Courses"].Value == null) return null;
                return group.Settings["Courses"].Value.ToString().Split('#'); 
            }
            set
            {
                string setval;

                if (value == null)
                    setval = "";
                else
                    setval = string.Join("#", value);

                //for (int i = 0; i < value.Length; i++)
                //{
                //    setval += value[i];

                //    if ((i + 1) < value.Length)
                //    {
                //        setval += "#";
                //    }
                //}


                group.Settings["Courses"].Value = setval;
                settings.Save();
            }
        }

        public static int AnnouncementsDateFilter
        {
            get 
            {
                if (group.Settings["AnnouncementsDateFilter"].Value == null) return 0;
                return (int)group.Settings["AnnouncementsDateFilter"].Value; 
            }
            set 
            {
                group.Settings["AnnouncementsDateFilter"].Value = value;
                settings.Save();
            }
        }
            

        public static bool ShowHints
        {
            get { return (bool)group.Settings["ShowHints"].Value; }
            set
            {
                group.Settings["ShowHints"].Value = value;
                settings.Save();
            }
        }

        public static bool RememberUsernameAndPassword
        {
            get { return (bool)group.Settings["RememberUsernameAndPassword"].Value; }
            set 
            {
                group.Settings["RememberUsernameAndPassword"].Value = value;
                settings.Save();
            }
        }

        public static string Username
        {
            get 
            { 
                return group.Settings["Username"].Value == null ? null : group.Settings["Username"].Value.ToString(); 
            }
            set 
            { 
                group.Settings["Username"].Value = value;
                settings.Save();
            }
        }

        // TODO: Use proper encryption?
        public static string Password
        {
            get
            {
                if (group.Settings["Password"].Value == null) return null;

                string base64val = group.Settings["Password"].Value.ToString();
                byte[] arr = Convert.FromBase64String(base64val);
                return UTF8Encoding.UTF8.GetString(arr, 0, arr.Length);
            }
            set
            {
                group.Settings["Password"].Value = string.IsNullOrEmpty(value) ? null : Convert.ToBase64String(UTF8Encoding.UTF8.GetBytes(value));
                settings.Save();
            }
        }

      

        public static CultureInfo Language
        {
            get
            {
                if (group.Settings["Language"].Value == null)
                {
                    return GetDefaultLanguage();
                }
                else
                {
                    return new CultureInfo(group.Settings["Language"].Value.ToString());
                }
                
            }
            set
            {
                group.Settings["Language"].Value = value.ToString();
                settings.Save();
            }
        }

        public static string UserServiceName
        {
            get { return "UserService.asmx"; }
        }
        public static string LocationServiceName
        {
            get { return "LocationService.asmx"; }
        }
        public static string ScheduleServiceName
        {
            get { return "ScheduleService.asmx"; }
        }
        public static string AnnouncementServiceName
        {
            get { return "AnnouncementService.asmx"; }
        }
        public static string LearningWebServiceName
        {
            get { return "LearningWebService.asmx"; }
        
        }

        public static string CollaborationWebserviceName 
        {
            get { return "CollaborationService.asmx"; }
        
        }

        public static string AppTitle
        {
            get { return "MCLS v1.2"; }
        }
    }
}