﻿using System;
using System.Linq;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;
using MSIS.webservices.location;

namespace MSIS
{
    public class MapSheet : Panel
    {
        enum MarkerType
        {
            Location = 0,
            FingerprintMarker = 1,
            GPSMarker = 2
        }

        private Image mapImage = null;
        private List<Location> locations = new List<Location>(); // Holds all locations shown on the map
        private int selectedLocationIndex = -1;
        private List<LocationPin> locationPins = new List<LocationPin>(); // Must be same size and order as 'locations'
        private ZoomButtons zoomButtons;
        private int zoomLevel = 0; // default zoom level
        private bool clicked;
        private int oldX;
        private int oldY;
        private int imgX;
        private int imgY;
        private int oldImgX; // used only for pin loc
        private int oldImgY; // --
        private int yOffset;
        private bool newMove = true;
        private int lineStartIdx = -1;
        private int lineEndIdx = -1;
        private bool showCircleAroundPin = true;
        private LocationFinder container;

        public MapSheet(LocationFinder container)
        {
            this.container = container;
            this.AutoScroll = false;
            this.MouseMove += new MouseEventHandler(PanImage_MouseMove);
            this.MouseDown += new MouseEventHandler(PanImage_MouseDown);
            this.MouseUp += new MouseEventHandler(PanImage_MouseUp);
            zoomButtons = new ZoomButtons(new Point(4, 4));
            this.Controls.Add(zoomButtons);
        }

        public bool AddLocation(Location location)
        {
            if (location.MapData.mapArray.Length < 1 || location.MapData.coordinateArray.Length < 1)
            {
                return false;
            }
            locations.Add(location);

            // convert coordinate string array into Point array
            string[] coordString = location.MapData.coordinateArray;
            Point[] coordinateArray = new Point[coordString.Length];
            for (int i = 0; i < coordString.Length; i++)
            {
                coordinateArray[i] = new Point(Int32.Parse(coordString[i].Split(',')[0]), Int32.Parse(coordString[i].Split(',')[1]));
            }

            // add pin on map
            LocationPin pin = new LocationPin(coordinateArray);
            pin.Location = coordinateArray[zoomLevel];
            locationPins.Add(pin);
            this.Controls.Add(pin);
            return true;
        }
        public bool RemoveLocation(Location location)
        {
            int idx = locations.IndexOf(location);
            this.Controls.Remove(locationPins[idx]);
            locationPins.RemoveAt(idx);
            return locations.Remove(location);
        }
        public void ClearLocations()
        {
            zoomLevel = 0;
            selectedLocationIndex = -1;
            lineStartIdx = lineEndIdx = -1;
            locations.Clear();
            // Remove location pins from map
            for (int i = 0; i < locationPins.Count; i++)
            {
                this.Controls.Remove(locationPins[i]);
            }
            locationPins.Clear();
        }

        public bool SetSelectedLocationIndex(int index)
        {
            if (index >= 0 && index < locations.Count)
            {
                selectedLocationIndex = index;
                if (container != null)
                {
                    container.ShowMapDescription(locations[selectedLocationIndex]);
                }
                return LoadMap(zoomLevel);
                //Parent.SetMapDescription(locations[selectedLocationIndex]);
            }
            return false;
        }

        public void DrawLine(Location start, Location end)
        {
            lineStartIdx = locations.IndexOf(start);
            lineEndIdx = locations.IndexOf(end);
        }

        public int NumberOfLocations
        {
            get { return locations.Count; }
        }

        public bool ShowCircleAroundPin
        {
            set { showCircleAroundPin = value; }
        }

        /// <summary>
        /// Sets the map sheet to display. Must be specified as file name of embedded resource (e.g.: map1.gif).
        /// </summary>
        /// <param name="mapName"></param>
        private bool LoadMap(int zoomLevel)
        {
            Cursor.Current = Cursors.WaitCursor;
            String[] mapArray = locations[selectedLocationIndex].MapData.mapArray;
            newMove = true;

            try
            {
                System.Net.WebRequest req = System.Net.WebRequest.Create(AppSettings.MapUrl + mapArray[zoomLevel]);
                System.Net.WebResponse resp = req.GetResponse();
                System.IO.Stream imgStream = resp.GetResponseStream();
                mapImage = new Bitmap(imgStream);
                imgStream.Close();
                resp.Close();
            }
            catch (Exception e)
            {
                return false;
            }

            // center map on pin
            imgX = imgY = 0;
            oldImgX = oldImgY = 0;
            int centerX = 0, centerY = 0;

            if (mapImage.Width > Width)
            {
                centerX = locationPins[selectedLocationIndex].CoordinateArray[zoomLevel].X - (Width / 2);
                if ((centerX + Width) > mapImage.Width)
                {
                    centerX = mapImage.Width - Width;
                }
                else if (imgX < 0 && (mapImage.Width > Width))
                {
                    centerX = 0;
                }
            }
            if (mapImage.Height > Height)
            {
                centerY = locationPins[selectedLocationIndex].CoordinateArray[zoomLevel].Y - (Height / 2);
                if ((centerY + Height) > mapImage.Height)
                {
                    centerY = mapImage.Height - Height;
                }
                else if (imgY < 0 && (mapImage.Height > Height))
                {
                    centerY = 0;
                }
            }

            imgX += centerX;
            imgY += centerY;
            oldImgX = imgX;
            oldImgY = imgY;


            for (int i = 0; i < locationPins.Count; i++)
            {
                // Remove location pins which are outside of this map,
                // e.g. does not have the same map as currently seleceted location at this zoomLevel
                locationPins[i].Enabled = false;
                if (zoomLevel < locations[i].MapData.mapArray.Length)
                {
                    locationPins[i].Enabled = (locations[i].MapData.mapArray[zoomLevel] == mapArray[zoomLevel]);
                }

                // update position of location pins according to zoom level
                if (locationPins[i].Enabled)
                {
                    locationPins[i].Location = locationPins[i].CoordinateArray[zoomLevel];
                    locationPins[i].Translate(centerX, centerY);
                }
            }

            Cursor.Current = Cursors.Default;
            return true;
        }

        private void ZoomIn()
        {
            String[] mapArray = locations[selectedLocationIndex].MapData.mapArray;
            if (zoomLevel < mapArray.Length - 1)
            {
                zoomLevel++;
                // load correct map according to zoom level
                if (LoadMap(zoomLevel))
                {
                    Refresh();
                    return;
                }
                zoomLevel--;
                MessageBox.Show("Unable to load map from server!","Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
            }
            else
            {
                MessageBox.Show("Maximum zoom level reached", "Info",
                    MessageBoxButtons.OK, MessageBoxIcon.Asterisk, MessageBoxDefaultButton.Button1);
            }
        }

        private void ZoomOut()
        {
            if (zoomLevel > 0)
            {
                zoomLevel--;
                // load correct map according to zoom level
                if (LoadMap(zoomLevel))
                {
                    Refresh();
                    return;
                }
                zoomLevel++;
                MessageBox.Show("Unable to load map from server!", "Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button1);
            }
            else
            {
                MessageBox.Show("Minimum zoom level reached", "Info",
                    MessageBoxButtons.OK, MessageBoxIcon.Asterisk, MessageBoxDefaultButton.Button1);
            }
        }

        private void PanImage_MouseDown(object Sender, MouseEventArgs e)
        {
            if (this.mapImage == null) return;

            // Hit-test for pins
            for (int i = 0; i < locationPins.Count; i++)
            {
                if (locationPins[i].Enabled)
                {
                    Rectangle rect = new Rectangle(locationPins[i].Bounds.X - locationPins[i].Bounds.Width,
                        locationPins[i].Bounds.Y - locationPins[i].Bounds.Height - 7,
                        locationPins[i].Bounds.Height + 10, locationPins[i].Bounds.Height + 10);
                    if (rect.Contains(e.X, e.Y))
                    {
                        selectedLocationIndex = i;
                        if (container != null)
                        {
                            container.ShowMapDescription(locations[selectedLocationIndex]);
                        }
                        Refresh();
                    }
                }
            }

            // Hit-test for zoom buttons
            if (e.X > zoomButtons.Location.X && e.X < (zoomButtons.Location.X + zoomButtons.Width / 2)
            && e.Y > zoomButtons.Location.Y && (e.Y < zoomButtons.Location.Y + zoomButtons.Height))
            {
                // + zoom in clicked
                ZoomIn();
            }
            else if (e.X > (zoomButtons.Location.X + zoomButtons.Width / 2) && e.X < (zoomButtons.Location.X + zoomButtons.Width)
                  && e.Y > zoomButtons.Location.Y && (e.Y < zoomButtons.Location.Y + zoomButtons.Height))
            {
                // - zoom out clicked
                ZoomOut();
            }

            else if (this.mapImage.Width > this.Width || this.mapImage.Height > this.Height)
            {
                if (newMove)
                {
                    for (int i = 0; i < locationPins.Count; i++)
                    {
                        if (locationPins[i].Enabled)
                        {
                            locationPins[i].OldLocation = new Point(locationPins[i].Location.X,
                                locationPins[i].Location.Y);
                        }
                    }
                    newMove = false;
                }

                // store current imgX, ImgY to use for calc. of pin positions later
                oldImgX = imgX;
                oldImgY = imgY;

                this.clicked = true;
                this.oldX = e.X;
                this.oldY = e.Y;
            }
        }
        private void PanImage_MouseMove(object Sender, MouseEventArgs e)
        {
            if (this.clicked)
            {
                int x = this.oldX - e.X;
                int y = this.oldY - e.Y;

                if (CalculateMove(x, y))
                {
                    this.Invalidate(this.ClientRectangle);
                }

                // update position of location pins in realtime.
                // position will be adjusted in MouseUp event.
                for (int i = 0; i < locationPins.Count; i++)
                {
                    LocationPin pin = locationPins[i];
                    if (pin.Enabled)
                    {
                        if (imgX > 0 && (imgX + Width < mapImage.Width) && oldImgX != imgX)
                        {
                            pin.Translate(x, 0);
                        }
                        if (imgY > 0 && (imgY + Height < mapImage.Height) && oldImgY != imgY)
                        {
                            pin.Translate(0, y);
                        }
                    }
                }

                this.oldX = e.X;
                this.oldY = e.Y;
            }
        }
        private void PanImage_MouseUp(object Sender, MouseEventArgs e)
        {
            if (this.Height > this.mapImage.Height)
            {
                imgY = oldImgY = 0;
            }

            // check if map has moved, if so update pin locations
            if (imgX != oldImgX || imgY != oldImgY)
            {
                for (int i = 0; i < locationPins.Count; i++)
                {
                    LocationPin pin = locationPins[i];
                    if (pin.Enabled)
                    {
                        pin.Location = new Point(pin.OldLocation.X - (imgX - oldImgX),
                                                 pin.OldLocation.Y - (imgY - oldImgY));
                    }
                }
                newMove = true;
                Refresh();
            }

            oldImgX = oldImgY = 0; // reset

            if (this.clicked)
            {
                this.clicked = false;
                this.oldX = this.oldY = 0;
            }
        }
        private bool CalculateMove(int x, int y)
        {
            if (x == 0 && y == 0) return false;
            this.imgY += y;
            this.imgX += x;
            if (this.imgY < 0) this.imgY = 0;
            if (this.imgX < 0) this.imgX = 0;
            if (this.imgY + this.Height > this.mapImage.Height) this.imgY = this.mapImage.Height - this.Height;
            if (this.imgX + this.Width > this.mapImage.Width) this.imgX = this.mapImage.Width - this.Width;


            if (this.Height > this.mapImage.Height) // TEST!!!
            {
                this.imgY = 0;
            }
            return true;
        }

        protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
        {
            if (Height > this.mapImage.Height)
            {
                yOffset = (Height / 2) - (this.mapImage.Height / 2);
            }
            else
            {
                yOffset = 0;
            }

            Rectangle destRect = new Rectangle(this.ClientRectangle.X, this.ClientRectangle.Y + yOffset, this.ClientRectangle.Width, this.ClientRectangle.Height);
            e.Graphics.FillRectangle(new SolidBrush(Color.White), this.ClientRectangle);
            e.Graphics.DrawImage(this.mapImage, destRect, new Rectangle(this.imgX, this.imgY, this.Width, this.Height), GraphicsUnit.Pixel);
            locationPins.ForEach(delegate(LocationPin pin)
            {
                if (pin.Enabled)
                {
                    pin.DrawPin(e.Graphics, yOffset);
                }
            });
            zoomButtons.Draw(e.Graphics);

            if (selectedLocationIndex != -1 && showCircleAroundPin)
            {
                Rectangle pinBounds = locationPins[selectedLocationIndex].Bounds;
                Rectangle rect = new Rectangle(pinBounds.X - pinBounds.Width, (pinBounds.Y - pinBounds.Height - 7) + yOffset, pinBounds.Height + 10, pinBounds.Height + 10);
                e.Graphics.DrawEllipse(new Pen(Color.Yellow, 4.0f), rect);
            }

            // Draw line between points
            if (lineStartIdx != lineEndIdx)
            {
                LocationPin pin1 = locationPins[lineStartIdx];
                LocationPin pin2 = locationPins[lineEndIdx];
                if (pin1.Enabled && pin2.Enabled)
                {
                    Point startPoint = pin1.Location;
                    Point endPoint = pin2.Location;
                    e.Graphics.DrawLine(new Pen(Color.Yellow, 4.0f),
                        startPoint.X, startPoint.Y, endPoint.X, endPoint.Y);
                }
            }
        }

        protected override void OnPaintBackground(PaintEventArgs e)
        {
            zoomButtons.Draw(e.Graphics);
        }
    }
}