using System;
using System.Drawing;
using System.IO;
using System.Reflection;

namespace MSIS
{
    /// <summary>
    /// Image that supports alpha blending / transparent background.
    /// Use with PNG image files that needs to be displayed on a surface where a transparent image background is desired.
    /// </summary>
    public sealed class AlphaImage
    {
        IImage _image;
        Size _size;

        /// <summary>
        /// Creates a new image based on the specified file path.
        /// </summary>
        /// <param name="path">The file path for the source image file.</param>
        /// <returns>An <see cref="AlphaImage"/> instance representing the specified image file.</returns>
        public static AlphaImage CreateFromFile(string path)
        {
            AlphaImage instance = new AlphaImage();
            instance.SetImage(CreateIImageFromFile(path));
            return instance;
        }

        /// <summary>
        /// Creates a new image based on an image resource in the specified assembly.
        /// </summary>
        /// <param name="assembly">The assembly containing the image resource to be used.</param>
        /// <param name="resourceName">The name of the image resource in the specified assembly's resource manifest.</param>
        /// <returns>An <see cref="AlphaImage"/> instance representing the specified assembly resource.</returns>
        public static AlphaImage CreateFromAssemblyResource(Assembly assembly, string resourceName)
        {
            AlphaImage instance = new AlphaImage();
            instance.SetImage(CreateIImageFromAssemblyResource(assembly, resourceName));
            return instance;
        }

        void SetImage(IImage image)
        {
            _image = image;
            ImageInfo imgInfo;
            _image.GetImageInfo(out imgInfo);
            _size = new Size((int)imgInfo.Width, (int)imgInfo.Height);
        }

        static IImagingFactory CreateIImagingFactory()
        {
            return (IImagingFactory)Activator.CreateInstance(Type.GetTypeFromCLSID(new Guid("327ABDA8-072B-11D3-9D7B-0000F81EF32E")));
        }

        static IImage CreateIImageFromFile(string path)
        {
            IImagingFactory factory = CreateIImagingFactory();
            IImage image;
            factory.CreateImageFromFile(path, out image);
            return image;
        }

        static IImage CreateIImageFromAssemblyResource(Assembly assembly, string resourceName)
        {
            IImagingFactory factory = CreateIImagingFactory();
            IImage image;
            MemoryStream resourceStream = (MemoryStream)assembly.GetManifestResourceStream(resourceName);
            // We must supply a byte buffer to the CreateImageFromBuffer method
            byte[] buffer = resourceStream.GetBuffer();
            factory.CreateImageFromBuffer(buffer, (uint)resourceStream.Length, BufferDisposalFlag.BufferDisposalFlagNone, out image);
            return image;
        }

        /// <summary>
        /// Draws the image on the specified <see cref="Graphics"/> object.
        /// </summary>
        /// <param name="surface">The target surface to draw the image on.</param>
        /// <param name="destination">The destination rectangle on the surface where the image is to be drawn.</param>
        public void Draw(Graphics surface, Rectangle destination)
        {
            // If we have an image, let it draw itself on the control's location on the form.
            IntPtr hdc = surface.GetHdc();
            _image.Draw(hdc, ref destination, IntPtr.Zero);
            surface.ReleaseHdc(hdc);
        }

        /// <summary>
        /// Returns the size of the image.
        /// </summary>
        public Size Size
        {
            get { return _size; }
        }
    }
}
