﻿using System;
using System.Data;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Services;
using System.Data.SqlClient;
using System.Drawing;
using System.Diagnostics;

/// <summary>
/// This web service provides the core functionality necessary for the location search service.
/// </summary>
[WebService(Namespace = "http://msis.myvnc.com/webservices/")]
[WebServiceBinding(ConformsTo = WsiProfiles.BasicProfile1_1)]
// To allow this Web Service to be called from script, using ASP.NET AJAX, uncomment the following line. 
[System.Web.Script.Services.ScriptService]
public class LocationService : System.Web.Services.WebService
{

    public LocationService () {

        //Uncomment the following line if using designed components 
        //InitializeComponent(); 
    }

    /// <summary>
    /// The Signal structure contains the SSID (Service Set Identifier), BSSID (Basic Service Set Identifier),
    /// and RSSI (Received Signal Strength Indication) of a wireless access point.
    /// </summary>
    public struct Signal
    {
        /// <summary>
        /// SSID of Access Point
        /// </summary>
        public string ssid;
        /// <summary>
        /// Unique identifier/MAC address of Access Point
        /// </summary>
        public string bssid;
        /// <summary>
        /// Signal strength value
        /// </summary>
        public int rssi;
        public override string ToString()
        {
            return "(" + bssid + "|" + rssi + ")";
        }
    }

    /// <summary>
    /// The Fingerprint structure contains a Signal vector and location ID. The ID refers to the
    /// unique identifier of a location in the database.
    /// </summary>
    public struct Fingerprint
    {
        /// <summary>
        /// Database record ID of the location associated with the fingerprint
        /// </summary>
        public int locationId;
        /// <summary>
        /// The signal vector associated with the fingerprint
        /// </summary>
        public Signal[] signal;

    }

    [WebMethod]
    [System.Web.Script.Services.ScriptMethod]
    public string[] GetLocationListAjax(string prefixText, int count)
    {
        Location[] matchingLocations;
        try
        {
            matchingLocations = GetLocationByName(prefixText);
        }
        catch (Exception ex)
        {
            return new string[] { "" };
        }
        List<String> locationNames = new List<string>();
        foreach (Location l in matchingLocations)
        {
            locationNames.Add(l.Name);
        }
        return locationNames.ToArray();
    }

    /// <summary>
    /// This method accepts a search string of arbitrary length and returns a list of locations matching
    /// the search string. The 'name', 'description', and 'keywords' fields associated with the location
    /// are included in the query.
    /// </summary>
    /// <param name="searchString">The search string</param>
    /// <returns>A list of matching locations or 'null' if no match is found</returns>
    [WebMethod]
    public Location[] GetLocationByName(string searchString)
    {
        string[] searchStrArray = searchString.Split(' ');
        List<Location> matchingLocations = null;

        string whereClause = "";
        for (int i = 0; i < searchStrArray.Length; i++)
        {
            whereClause += "(name LIKE '%" + searchStrArray[i] + "%' OR description LIKE '%" + searchStrArray[i] + "%' OR keywords LIKE '%" + searchStrArray[i] + "%')";
            if (i < searchStrArray.Length - 1)
            {
                whereClause += " AND ";
            }
        }
        string sqlQuery = "SELECT * FROM Locations WHERE " + whereClause;

        if (!SQLConn.Open())
        {
            return null;
        }

        SqlCommand cmd = SQLConn.ExecuteCommand(sqlQuery);
        if (cmd != null)
        {
            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                matchingLocations = new List<Location>();

                while (reader.Read())
                {
                    matchingLocations.Add(ConstructLocation(reader));
                }
                reader.Close();
            }
        }
        SQLConn.Close();
        return (matchingLocations != null && matchingLocations.Count > 0) ? matchingLocations.ToArray() : null;
    }

    /// <summary>
    /// This method queries the database for locations with fingerprints matching the characteristics of a
    /// given signal vector (usually the observed signal), in order to add location-awareness to the
    /// application.
    /// </summary>
    /// <param name="signalVector">An array of signals to use for the lookup</param>
    /// <returns>The location if a match is found, otherwise 'null'</returns>
    [WebMethod]
    public Location GetLocationByFingerprint(Signal[] signalVector)
    {
        Location matchingLocation = null;

        // select all fingerprints which include one of the access points (bssid's)
        string sqlQuery =
        "SELECT Fingerprints.loc_id as LocationID, Fingerprints.rssi as RSSI, WIFI_APS.bssid as BSSID, WIFI_APS.ssid as SSID " +
        "FROM Fingerprints " +
        "INNER JOIN Locations ON Fingerprints.loc_id = Locations.id " +
        "INNER JOIN WIFI_APS ON Fingerprints.ap_id = WIFI_APS.id ";

        string whereClause = "WHERE ";
        for (int i = 0; i < signalVector.Length; i++)
        {
            if (i < (signalVector.Length - 1))
            {
                whereClause += "WIFI_APS.bssid='" + signalVector[i].bssid + "' OR ";
            }
            else
            {
                whereClause += "WIFI_APS.bssid='" + signalVector[i].bssid + "'";
            }
        }

        sqlQuery += whereClause += " ORDER BY Fingerprints.loc_id, Fingerprints.rssi DESC";

        if (!SQLConn.Open())
        {
            return null; // Unable to connect to DB server, return null
        }

        SqlCommand cmd = SQLConn.ExecuteCommand(sqlQuery);

        bool resultsFound = false;
        List<Fingerprint> candidateFingerprints = new List<Fingerprint>();

        if (cmd != null)
        {
            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                // Get list of candidate fingerprints

                int lastLocationId = 0;
                Fingerprint fp = new Fingerprint();
                int signalCounter = 0;

                while (reader.Read())
                {
                    int currentLocationId = Int32.Parse(reader["LocationID"].ToString());
                    if (lastLocationId != currentLocationId)
                    {
                        if (lastLocationId != 0)
                        {
                            // the previous fingerprint is complete -> add to list
                            candidateFingerprints.Add(fp);
                        }
                        // prepare a new fingerprint
                        fp = new Fingerprint();
                        fp.locationId = currentLocationId;
                        fp.signal = new Signal[3];
                        signalCounter = 0; // reset counter
                        lastLocationId = currentLocationId;
                        resultsFound = true;
                    }

                    fp.signal[signalCounter].ssid = (string)reader["SSID"];
                    fp.signal[signalCounter].bssid = (string)reader["BSSID"];
                    fp.signal[signalCounter].rssi = Int32.Parse(reader["RSSI"].ToString());
                    signalCounter++;

                }
                if (resultsFound)
                {
                    // add the last fingerprint
                    candidateFingerprints.Add(fp);
                }

                reader.Close();
            }
        }
      

        if (resultsFound)
        {           
            Fingerprint targetFingerprint = MatchFingerprints(signalVector, candidateFingerprints);

            cmd = SQLConn.ExecuteCommand("SELECT * FROM Locations WHERE id=" + targetFingerprint.locationId);
            if (cmd != null)
            {
                using (SqlDataReader result = cmd.ExecuteReader())
                {
                    if (result.Read())
                    {
                        matchingLocation = ConstructLocation(result);
                    }
                    result.Close();
                }
            }
        }
        SQLConn.Close();

        return (matchingLocation != null) ? matchingLocation : null;
    }

    /// <summary>
    /// This method can be used to add location-fingerprint mapppings to the database. The method accept 
    /// a Location object with information about the location, and the corresponding signal vector which 
    /// constitutes the fingerprint. Existing access points are discovered and rejected to prevent
    /// duplicate database entries.
    /// </summary>
    /// <param name="location">The location object</param>
    /// <param name="signalVector">The signal vector</param>
    /// <returns>'True' if the location was successfully added, 'False' otherwise</returns>
 
    [WebMethod]
    public bool AddLocation(Location location, Signal[] signalVector)
    {
        if (!SQLConn.Open())
        {
            return false;
        }

        string coordinates = "";
        if (location.MapData.coordinateArray != null || location.MapData.coordinateArray.Length > 0)
        {
            coordinates = string.Join("|", location.MapData.coordinateArray);
        }
        
        string mapstring = "";
        if (location.MapData.mapArray != null && location.MapData.mapScale != null && location.MapData.mapArray.Length > 0 && location.MapData.mapScale.Length > 0)
        {
            string[] tmpArr = new string[location.MapData.mapArray.Length];

            for (int i = 0; i < location.MapData.mapArray.Length; i++)
            {
                tmpArr[i] = location.MapData.mapArray[i] + (location.MapData.mapScale[i] == 0 ? "" : "#" + location.MapData.mapScale[i]);
            }

            mapstring = string.Join(",", tmpArr);
        }


        SqlCommand cmd = null;
        cmd = SQLConn.ExecuteCommand("INSERT INTO Locations (name, description, keywords, floor, type, map_coordinates, map_array) " +
                "values (@locName, @locDesc, @locKeywords, @locFloor, @locType, @locCoordinates, @locMapArray)");
        cmd.Parameters.AddWithValue("locName", location.Name);
        cmd.Parameters.AddWithValue("locDesc", location.Description);
        cmd.Parameters.AddWithValue("locKeywords", location.Keywords);
        cmd.Parameters.AddWithValue("locFloor", location.Floor);
        cmd.Parameters.AddWithValue("locType", (int)location.LocationType);
        cmd.Parameters.AddWithValue("locCoordinates", coordinates);
        cmd.Parameters.AddWithValue("locMapArray", mapstring); 

        //cmd = SQLConn.ExecuteCommand("INSERT INTO Locations (name, description, keywords, floor, type, map_coordinates, map_array) " +
        //        "values ('" + location.Name + "', '" + location.Description + "','" + location.Keywords + "', 0, " + (int)location.LocationType + ", '', '')");

        bool locInsertSuccess = (cmd.ExecuteNonQuery() > 0);
        int locationID = 0;
        if (locInsertSuccess)
        {   
            // get ID of last inserted record
            locationID = Int32.Parse(SQLConn.ExecuteCommand("SELECT @@IDENTITY").ExecuteScalar().ToString());
        }
        
        bool apsInsertSuccess = true;
        bool fingerprintInsertSuccess = true;
        foreach (Signal signal in signalVector)
        {
            // check first if AP exists in DB, if not insert
            object apid = SQLConn.ExecuteCommand("SELECT id FROM WIFI_APS WHERE bssid='" + signal.bssid + "'").ExecuteScalar();
            int AccessPointID;
            if (apid == null) // AP is new, proceed with insert
            {
                cmd = SQLConn.ExecuteCommand("INSERT INTO WIFI_APS (bssid, ssid) " +
                        "values ('" + signal.bssid + "', '" + signal.ssid + "')");
                if (cmd.ExecuteNonQuery() == 0)
                {
                    apsInsertSuccess = false;
                }
                AccessPointID = Int32.Parse(SQLConn.ExecuteCommand("SELECT @@IDENTITY").ExecuteScalar().ToString());
            }
            else // AP exists, use ID
            {
                AccessPointID = Int32.Parse(apid.ToString());
            }
            if (apsInsertSuccess)
            {
                // construct Fingerprint by joining Location and APs
                cmd = SQLConn.ExecuteCommand("INSERT INTO Fingerprints (loc_id, ap_id, rssi) " +
                        "values (" + locationID + ", " + AccessPointID + ", " + signal.rssi + ")");
                if (cmd.ExecuteNonQuery() == 0)
                {
                    fingerprintInsertSuccess = false;
                }
            }
        }
        SQLConn.Close();
        return (locInsertSuccess && apsInsertSuccess && fingerprintInsertSuccess);
    }

    /// <summary>
    /// This method returns a list of locations in close proximity to the position where a given signal
    /// vector was observed. The list of locations are sorted in descending order according to their distance from the signal.
    /// </summary>
    /// <param name="locationType">The type of locations to search for</param>
    /// <param name="signalVector">The signal vector</param>
    /// <param name="locationsToReturn">The number of locations to return</param>
    /// <returns>The list of matching locations, or 'null' in case no match is found</returns>
   
    
    [WebMethod]
    public Location[] GetNearbyLocations(Location.LocationTypes locationType, Signal[] signalVector, int locationsToReturn)
    {
        // TODO: Query database for locations matching 'locationType', and that are closest to
        // current location (based on signalVector). Return 'locationsToReturn' locations.
        Location currentLocation = GetLocationByFingerprint(signalVector);
        List<Location> matchingLocations = null;

        if (!SQLConn.Open())
        {
            return null;
        }

        // Retrieve all locations of type 'locationType'
        SqlCommand cmd = SQLConn.ExecuteCommand("SELECT * FROM Locations WHERE type=" + (int)locationType);
        if (cmd != null)
        {
            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                matchingLocations = new List<Location>();
                while (reader.Read())
                {
                    Location location = ConstructLocation(reader); // location at this iteration

                    // Determine maximum zoom level where locations have same map.
                    // We only need to iterate through the smallest of the two mapArrays
                    int smallestMapArray = (currentLocation.MapData.mapArray.Length > location.MapData.mapArray.Length)
                                         ? location.MapData.mapArray.Length
                                         : currentLocation.MapData.mapArray.Length;

                    int zoomLevel = -1;
                    for (int i = 0; i < smallestMapArray; i++)
                    {
                        if (currentLocation.MapData.mapArray[i] == location.MapData.mapArray[i])
                        {
                            zoomLevel++;
                        }
                    }

                    if (zoomLevel > -1)
                    {
                        Point currentPosition = GetCoordinatePoint(currentLocation, zoomLevel);
                        Point locationPosition = GetCoordinatePoint(location, zoomLevel);

                        // Adjust distance according to map scale and zoom level
                        if (location.MapData.mapScale[zoomLevel] > 0)
                        {
                            location.DistanceFromCurrent = EuclideanDistance(currentPosition, locationPosition) / 50 * location.MapData.mapScale[zoomLevel];
                        }
                    }
                    
                    matchingLocations.Add(location);
                }
                reader.Close();
            }
        }
        SQLConn.Close();

        if (matchingLocations != null && matchingLocations.Count > 0)
        {
            // Sort locations according to distance from currentPosition
            matchingLocations.Sort(delegate(Location l1, Location l2)
            {
                if (l1.DistanceFromCurrent < 0 && l2.DistanceFromCurrent >=0)
                {
                    return 1;
                }
                else if (l1.DistanceFromCurrent >= 0 && l2.DistanceFromCurrent < 0)
                {
                    return -1;
                }
                else if (l1.DistanceFromCurrent == l2.DistanceFromCurrent)
                {
                    return 0;
                }
                else if (l1.DistanceFromCurrent < l2.DistanceFromCurrent)
                {
                    return -1;
                }
                else
                {
                    return 1;
                }
            });
            return (locationsToReturn > 0) ? matchingLocations.Take(locationsToReturn).ToArray() : matchingLocations.ToArray();
        }
        return null;
    }

    /// <summary>
    /// Builds Location object from SqlDataReader row.
    /// </summary>
    /// <param name="reader">The data reader from 'Location' table</param>
    /// <param name="distFromCurrentPos">Optional distance from current position. Use -1 for no value.</param>
    /// <returns>The Location object</returns>
    private Location ConstructLocation(SqlDataReader reader)
    {
        Location location = new Location();
        location.Name = (string)reader["name"];
        location.Keywords = (string)reader["keywords"];
        location.Description = ((string)reader["description"] == string.Empty) ? "No description is available." : (string)reader["description"];
        location.Floor = int.Parse(reader["floor"].ToString());
        location.LocationType = (Location.LocationTypes)int.Parse(reader["type"].ToString());

        string[] coordinateArray = new string[0];
        List<string> mapArray = new List<string>();
        List<int> mapScales = new List<int>();
        if (reader["map_coordinates"].ToString() != "" && reader["map_array"].ToString() != "")
        {
            // Map coordinates
            coordinateArray = reader["map_coordinates"].ToString().Split('|');
            // Map array
            string[] tmpMapString = reader["map_array"].ToString().Split(',');

            for (int i = 0; i < tmpMapString.Length; i++)
            {
                string[] tmp2 = tmpMapString[i].Split('#');
                mapArray.Add(tmp2[0]);
                mapScales.Add((tmp2.Length > 1) ? Int32.Parse(tmp2[1]) : 0);
            }
        }
        location.MapData = new Location.MapStruct(coordinateArray, mapArray.ToArray(), mapScales.ToArray());

        return location;
    }

    /// <summary>
    /// Returns the fingerprint in a list of candidate fingerprints matching the characteristics of 
    /// the observed signal. If two identical matches are found, the first match is returned.
    /// </summary>
    /// <param name="subject">The observed RSSI fingerprint</param>
    /// <param name="candidates">A list of possible candidate fingerprints</param>
    /// <returns></returns>
    private Fingerprint MatchFingerprints(Signal[] observedSignal, List<Fingerprint> candidates)
    {
        double closestMatchValue = 0.0;
        int closestMatchIdx = 0;
        double d;

        for (int i = 0; i < candidates.Count; i++) // for each fingerprint
        {
            Signal[] candidateSignalVector = candidates[i].signal;
            
            d = EuclideanDistance(observedSignal, candidateSignalVector);
            if (d < closestMatchValue || i == 0)
            {
                closestMatchValue = d;
                closestMatchIdx = i;
            }
        }
        return candidates[closestMatchIdx];
    }

    private double EuclideanDistance(Signal[] observedSignal, Signal[] signalVector)
    {
        // TODO: include observered signal factors in terms although not contained in signalvector
        double euclideanDistance = 0;

        foreach (Signal signal in observedSignal)
        {
            bool bssidFound = false;
            for (int i = 0; i < signalVector.Length; i++)
            {
                if (signalVector[i].bssid == signal.bssid) // true at most once per vector
                {
                    euclideanDistance += Math.Pow(signal.rssi - signalVector[i].rssi, 2);
                    bssidFound = true;
                }
            }
            if (!bssidFound) // observed AP/bssid not in fingerprint
            {
                // add "penalty" for missing AP
                euclideanDistance += Math.Pow(65, 2); //Math.Pow(signal.rssi, 2);
            }
        }
        return Math.Sqrt(euclideanDistance);
    }
    /// <summary>
    /// Calculates the distance between two points.
    /// </summary>
    /// <param name="p1">The first point</param>
    /// <param name="p2">The second point</param>
    /// <returns>The distance.</returns>
    private double EuclideanDistance(Point p1, Point p2)
    {
        Point invalid = new Point(-1, -1);
        if (Point.Equals(p1, invalid) || Point.Equals(p2, invalid))
        {
            // Point not set
            return -1;
        }
        return Math.Sqrt(Math.Pow(p2.X - p1.X, 2) + Math.Pow(p2.Y - p1.Y, 2));
    }

    /// <summary>
    /// Returns the Point representation of the coordinate pair at a given zoom level
    /// </summary>
    /// <param name="location">The location</param>
    /// <param name="level">The zoom level whos point to return</param>
    /// <returns>The Point corresponding to the coordinate pair</returns>
    private Point GetCoordinatePoint(Location location,int level)
    {
        Point p = new Point(-1,-1);
        if (level < location.MapData.coordinateArray.Length && level >= 0)
        {
            string[] coordinates = location.MapData.coordinateArray[level].Split(',');
            p = new Point(int.Parse(coordinates[0]), int.Parse(coordinates[1]));
        }
        return p;
    }

    


    private void InitializeComponent()
    {


    }

    [WebMethod]
    public void SaveCurrentLocation(string UserId, int LocationId)
    {
        int count = 0;
        SQLConn.Open();
        using (SqlCommand cm = SQLConn.ExecuteCommand("SELECT Count(*) FROM User_Location WHERE User_id = " + UserId ))
        {
            using (SqlDataReader result = cm.ExecuteReader())
            {
                if (result.Read())
                    count = Convert.ToInt32(result[0]);
                result.Close();
            }

        }

        string strSql = null;

        if (count > 0)
        {
            strSql = "UPDATE User_Location SET location_id = " + LocationId + ",date='" +
                DateTime.Now.ToString() + "' WHERE user_id = " + UserId;
            
        }
        else
        {
            strSql = "INSERT INTO User_Location(user_id, location_id, date) VALUES(" + UserId.ToString() + ", " +
                LocationId.ToString() + ", '" + DateTime.Now.ToString() + "')";
        }

        SqlCommand cmd = SQLConn.ExecuteCommand(strSql);
        cmd.ExecuteNonQuery();                
   }



[WebMethod]
   public string[] GetNearByUsers(int LocationId)
   {
        //Init Connection string etc
       List<string> userIds = new List<string>();      
                  
         SQLConn.Open();
         using (SqlCommand cm = SQLConn.ExecuteCommand("SELECT U.name FROM Users u WHERE u.id IN (SELECT user_id FROM User_Location where location_id = " + LocationId + ")"))
        {            

            using (SqlDataReader result = cm.ExecuteReader())
            {
               
                // If execution gives a result:
               
                while (result.Read())
                {


                    userIds.Add(result["name"].ToString());
                   
                }
                result.Close();
            }


         }
       return userIds.ToArray();
   }

  /*  [WebMethod]
    public Location GetUserLocation(int UserId)
    {
        Location userLoc = null;
        SQLConn.Open();
        using (SqlCommand cm = SQLConn.ExecuteCommand("select * from Locations WHERE id = (SELECT location_id FROM User_Location WHERE user_id = " + UserId + ")"))
        {
            using (SqlDataReader result = cm.ExecuteReader())
            {
                while (result.Read())
                {
                    userLoc = new Location();
                    userLoc.Name = result["name"].ToString();
                    userLoc.Description = result["description"].ToString();
                }

                result.Close();
            }
        }

        return userLoc;
   }*/

    [WebMethod]
    public Location[] GetUserLocation(string Name)
    {        
        List<Location> userLoc = new List<Location>();
        SQLConn.Open();
        using (SqlCommand cm = SQLConn.ExecuteCommand("select * from Locations WHERE id IN (SELECT location_id FROM User_Location WHERE user_id IN( SELECT id from Users where name like '%" + Name + "%' ))"))
        {
            using (SqlDataReader result = cm.ExecuteReader())
            {
                while (result.Read())
                {
                    Location l = new Location();
                    l.Name = result["name"].ToString();
                    l.Description = result["description"].ToString();
                    userLoc.Add(l);
                }

                result.Close();
            }
        }

        return userLoc.ToArray();
    }


    [WebMethod]
    public Location[] GetUsers(string LocationName)
    {
        //Init Connection string etc
        List<Location> userName = new List<Location>();

        SQLConn.Open();
        using (SqlCommand cm = SQLConn.ExecuteCommand("select * from Users where id in (select user_id from User_Location where location_id in(SELECT id from Locations where name like '%" + LocationName + "%'))"))
        {

            using (SqlDataReader result = cm.ExecuteReader())
            {

                // If execution gives a result:

                while (result.Read())
                {
                    Location gUsers = new Location();

                    gUsers.Name = result["name"].ToString();
                   // userName.Add(result["name"].ToString());
                    userName.Add(gUsers);
                }
                result.Close();
            }


        }
        return userName.ToArray();
    }





}


